#!/usr/bin/env python
# -*- coding: utf-8 -*-


from .activation import *
from .crop import *
from .get import *
from .indices import *
from .mask import *
from .multilabel import *
from .others import *
from .pad import *
from .repeat import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os


def get_auto_num_cpus() -> int:
    """Returns the number of CPUs available, useful for num_workers attribute used in DataLoaders."""
    return len(os.sched_getaffinity(0))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

T = TypeVar("T")
U = TypeVar("U")


def list_dict_to_dict_list(
    lst: Sequence[Mapping[str, T]],
    default_val: U = None,
    error_on_missing_key: bool = False,
) -> Dict[str, List[Union[T, U]]]:
    """Convert a list of dicts to a dict of lists.

    Example 1
    ----------
    >>> lst = [{'a': 1, 'b': 2}, {'a': 4, 'b': 3, 'c': 5}]
    >>> output = list_dict_to_dict_list(lst, default_val=0)
    {'a': [1, 4], 'b': [2, 3], 'c': [0, 5]}
    """
    if len(lst) == 0:
        return {}

    if error_on_missing_key:
        keys = set(lst[0])
        for dic in lst:
            if keys != set(dic.keys()):
                raise ValueError(
                    f"Invalid dict keys for list_dict_to_dict_list. (found {keys} and {dic.keys()})"
                )

    keys = {}
    for dic in lst:
        keys.update(dict.fromkeys(dic.keys()))

    out = {
        key: [
            lst[i][key] if key in lst[i].keys() else default_val
            for i in range(len(lst))
        ]
        for key in keys
    }
    return out


def dump_dict(
    dic: Mapping[str, Any],
    /,
    join: str = ", ",
    fmt: str = "{name}={value}",
    ignore_none: bool = False,
) -> str:
    result = join.join(
        fmt.format(name=name, value=value)
        for name, value in dic.items()
        if not (value is None and ignore_none)
    )
    return result


def pass_filter(
    name: T,
    include: Optional[Iterable[T]] = None,
    exclude: Optional[Iterable[T]] = None,
) -> bool:
    """Returns True if name in include set and not in exclude set."""
    if include is not None and exclude is not None:
        return (name in include) and (name not in exclude)
    if include is not None:
        return name in include
    elif exclude is not None:
        return name not in exclude
    else:
        return True


def filter_iterable(
    it: Iterable[T],
    include: Optional[Iterable[T]] = None,
    exclude: Optional[Iterable[T]] = None,
) -> List[T]:
    return [item for item in it if pass_filter(item, include, exclude)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import torch
from torch import Tensor, nn

from torchoutil.utils.packaging import _NUMPY_AVAILABLE

if _NUMPY_AVAILABLE:
    import numpy as np

    class ToNumpy(nn.Module):
        def forward(self, x: Tensor) -> np.ndarray:
            return x.cpu().numpy()

    class FromNumpy(nn.Module):
        def forward(self, x: np.ndarray) -> Tensor:
            return torch.from_numpy(x)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os.path as osp
from typing import Any, Dict, List

from torchoutil.utils.packaging import _TENSORBOARD_AVAILABLE

pylog = logging.getLogger(__name__)


EVENT_FILE_PREFIX = "events.out.tfevents."
DT_FLOAT = 1
DT_STRING = 7


if _TENSORBOARD_AVAILABLE:
    from tensorboard.backend.event_processing.event_file_loader import (  # type: ignore
        EventFileLoader,
    )

    def load_event_file(
        fpath: str,
        cast_float_and_str: bool = True,
        ignore_underscore_tags: bool = True,
        verbose: int = 0,
    ) -> List[Dict[str, Any]]:
        if not osp.isfile(fpath):
            raise FileNotFoundError(f"Invalid argument {fpath=}. (not a file)")

        event_file_loader = EventFileLoader(fpath)
        raw_data = []

        for event in event_file_loader.Load():
            wall_time: float = event.wall_time  # type: ignore
            event_values: list = event.summary.value  # type: ignore
            step: int = event.step  # type: ignore

            for event_value in event_values:
                tag = event_value.tag
                dtype = event_value.tensor.dtype
                string_val = event_value.tensor.string_val
                float_val = event_value.tensor.float_val

                data_i = {
                    "wall_time": wall_time,
                    "step": step,
                    "tag": tag,
                    "dtype": dtype,
                    "string_val": string_val,
                    "float_val": float_val,
                }
                raw_data.append(data_i)

        data = []
        for data_i in raw_data:
            tag: str = data_i["tag"]
            dtype: Any = data_i["dtype"]

            if ignore_underscore_tags and tag.startswith("_"):
                if verbose >= 2:
                    pylog.debug(
                        f'Skip value with tag "{tag}" which begins by an underscore.'
                    )
                continue

            if cast_float_and_str:
                if dtype == DT_FLOAT:
                    float_val: List[float] = data_i["float_val"]
                    assert len(float_val) == 1
                    value = float_val[0]
                    dtype = "float"

                elif dtype == DT_STRING:
                    string_val: str = data_i["string_val"]
                    value = string_val[3:-2]
                    dtype = "str"
                    tag = tag.split("/")[0]

                else:
                    raise RuntimeError(f"Unknown value {dtype=}.")

                del data_i["string_val"]
                del data_i["float_val"]

                data_i["tag"] = tag
                data_i["dtype"] = dtype
                data_i["value"] = value

            data.append(data_i)

        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Iterable

from torch import Tensor


def softmax_multidim(
    input: Tensor,
    dims: Iterable[int] = (-1,),
) -> Tensor:
    """A multi-dimensional version of torch.softmax."""
    dims = tuple(dims)
    x = input.exp()
    result = x / x.sum(dim=dims, keepdim=True)
    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Generic, OrderedDict, TypeVar, overload

from torch import nn

InType = TypeVar("InType", covariant=False, contravariant=True)
OutType = TypeVar("OutType", covariant=True, contravariant=False)
InterType1 = TypeVar("InterType1")
InterType2 = TypeVar("InterType2")
InterType3 = TypeVar("InterType3")
InterType4 = TypeVar("InterType4")
InterType5 = TypeVar("InterType5")
InterType6 = TypeVar("InterType6")


class TModule(nn.Module, Generic[InType, OutType]):
    """Typed version of torch.nn.Module. Can specify an input and output type."""

    def __call__(self, *args: InType, **kwargs: InType) -> OutType:
        return super().__call__(*args, **kwargs)

    def forward(self, *args: InType, **kwargs: InType) -> OutType:
        return super().forward(*args, **kwargs)


class TSequential(nn.Sequential, TModule, Generic[InType, OutType]):
    """Typed version of torch.nn.Sequential, designed to work with torchoutil.nn.TModules."""

    @overload
    def __init__(
        self,
        arg0: TModule[InType, OutType],
        /,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TModule[InType, InterType1],
        arg1: TModule[InterType1, OutType],
        /,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TModule[InType, InterType1],
        arg1: TModule[InterType1, InterType2],
        arg2: TModule[InterType2, OutType],
        /,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TModule[InType, InterType1],
        arg1: TModule[InterType1, InterType2],
        arg2: TModule[InterType2, InterType3],
        arg3: TModule[InterType3, OutType],
        /,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TModule[InType, InterType1],
        arg1: TModule[InterType1, InterType2],
        arg2: TModule[InterType2, InterType3],
        arg3: TModule[InterType3, InterType4],
        arg4: TModule[InterType4, OutType],
        /,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TModule[InType, InterType1],
        arg1: TModule[InterType1, InterType2],
        arg2: TModule[InterType2, InterType3],
        arg3: TModule[InterType3, InterType4],
        arg4: TModule[InterType4, InterType5],
        arg5: TModule[InterType5, OutType],
        /,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg0: TModule[InType, InterType1],
        arg1: TModule[InterType1, InterType2],
        arg2: TModule[InterType2, InterType3],
        arg3: TModule[InterType3, InterType4],
        arg4: TModule[InterType4, InterType5],
        arg5: TModule[InterType5, InterType6],
        arg6: TModule[InterType6, OutType],
        /,
    ) -> None:
        ...

    @overload
    def __init__(self, *args: nn.Module) -> None:
        ...

    @overload
    def __init__(self, arg: "OrderedDict[str, nn.Module]") -> None:
        ...

    @overload
    def __init__(self, arg: "OrderedDict[str, TModule[InType, OutType]]") -> None:
        ...

    def __init__(self, *args) -> None:  # type: ignore
        nn.Sequential.__init__(self, *args)
        TModule.__init__(self)

    def __call__(self, *args: InType, **kwargs: InType) -> OutType:
        return nn.Sequential.__call__(self, *args, **kwargs)

    def forward(self, *args: InType, **kwargs: InType) -> OutType:
        return nn.Sequential.forward(self, *args, **kwargs)


def __test_typing() -> None:
    import torch
    from torch import Tensor

    class LayerA(nn.Module):
        def forward(self, x: Tensor) -> Tensor:
            return x * x

    class LayerB(nn.Module):
        def forward(self, x: Tensor) -> int:
            return int(x.sum().item())

    x = torch.rand(10)
    xa = LayerA()(x)
    xb = LayerB()(x)

    seq = TSequential(LayerA(), LayerA(), LayerB())
    xab = seq(x)

    assert isinstance(xa, Tensor)
    assert isinstance(xb, int)
    assert isinstance(xab, int)
